/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <joerg@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.        Joerg Wunsch
 * ----------------------------------------------------------------------------
 */

#include <stdbool.h>

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>

#include "timer.h"

/*
 * Bitmask for USB9Out bits that are considered to be active.
 */
#define USB9OUT_MASK 0x0F

/*
 * Bitmask for external device on port C
 */
#define EXT_DEVICE   0x04

/*
 * Bitmask for emergency poweroff on port C
 */
#define POWER_OFF    0x02

/*
 * Bitmask for output switch, active low
 */
#define OUTPUT_SWITCH 0x01

static volatile bool timer_triggered;
static volatile bool pcint_triggered;

static uint8_t portd_mirrored;

static void *handle_500ms;

static enum
{
    NOT_READY, STARTING_UP, READY
}
external_state;

ISR(TIMER1_COMPA_vect)
{
    timer_triggered = true;
}

ISR(PCINT2_vect)
{
    pcint_triggered = true;
    portd_mirrored = PIND & USB9OUT_MASK;
}

static void
ioinit(void)
{
    /*
     * timer 1 at full speed (1 MHz default clock), CTC mode with
     * compare1A = 1000 => 1 ms timer interrupt service
     */
    OCR1A = 1000;
    TCCR1B = _BV(WGM12) /* CTC mode */ | _BV(CS10) /* clock source */;

    /*
     * Setup port B:
     * PB0...PB2: unused, internal pullup (just to keep them silent)
     * PB3...PB5: ISP, internal pullup (just to keep them silent)
     */
    PORTB = 0x3F;

    /*
     * Setup port C:
     * PC0: output to driver, low active, default to high
     * PC1: external switch to GND, internal pullup, turn off device when low
     * PC2: external ready signal, internal pullup, start activity when high for >= 6 s
     * PC3...PC5: unused, internal pullup (just to keep them silent)
     */
    PORTC = 0x3F;
    DDRC = OUTPUT_SWITCH;

    /*
     * Setup port D:
     * PD0...PD7 input, digital inputs from USB9Out
     *
     * No action needed.
     */

    /* enable pin-change interrupts PCINT16...20 */
    PCMSK2 = USB9OUT_MASK;
    PCICR = _BV(PCIE2);

    sei();
}

/*
 * Timer callbacks.
 */

/*
 * 6-second delay handler for startup after external device on PC2
 * became ready.
 */
static void
startup_delay_done(union timeoutarg dummy)
{
    if (external_state == STARTING_UP)
    {
        external_state = READY;
    }
}

/*
 * Worker handler to poll for external device and external
 * radio-controlled switch.  Called each 10 ms.
 */
static void
maintenance_10ms(union timeoutarg dummy)
{
    if (PINC & EXT_DEVICE)
    {
        /*
         * external device ready, activate startup timer if necessary
         */
        if (external_state == NOT_READY)
        {
            external_state = STARTING_UP;
            timeout(6000, startup_delay_done, dummy);
        }
    }
    else
    {
        /*
         * external device not ready, power everything down
         */
        PORTC |= OUTPUT_SWITCH;
        external_state = NOT_READY;
    }

    if ((PINC & POWER_OFF) == 0)
    {
        /* remote power off */
        PORTC |= OUTPUT_SWITCH;
    }

    timeout(10, maintenance_10ms, dummy);
}

/*
 * Worker handler to extend USB9Out ready condition to 500 ms.
 */
static void
delayed_500ms(union timeoutarg dummy)
{
    /* 500 ms timer expired, turn off */
    PORTC |= OUTPUT_SWITCH;
}

/*
 * High-level pin-change interrupt handler.  Monitor pins, and if
 * all are set (according to mask), activate output, and fire 500 ms
 * delay.
 */
static void
pinchange_handler(void)
{
    union timeoutarg dummy = { .i = 0 };

    if (external_state != READY)
        /* don't do anything */
        return;

    if ((PINC & POWER_OFF) == 0)
        /* turned off by remote control */
        return;

    if (portd_mirrored == USB9OUT_MASK)
    {
        /* fine, all our inputs are ready */
        if (handle_500ms)
            untimeout(handle_500ms);

        handle_500ms = timeout(500, delayed_500ms, dummy);

        /* Go! */
        PORTC &= ~OUTPUT_SWITCH;
    }
}

int
main(void)
{
    union timeoutarg dummy = { .i = 0 };

    ioinit();
    timeout(10, maintenance_10ms, dummy);

    for (;;)
    {
        if (timer_triggered)
        {
            timer_triggered = false;
            timertick();
        }

        if (pcint_triggered)
        {
            pcint_triggered = false;
            pinchange_handler();
        }

        sleep_mode();
    }
}
